@@ -10,18 +10,39 @@ import os
 import re
 import string
 import sys
-from time import monotonic
 from contextlib import suppress
+from functools import lru_cache
+from time import monotonic
+from typing import (
+    Any, Callable, Dict, Generator, Iterable, List, Match, NamedTuple,
+    Optional, Tuple, Union, cast
+)
 
 from .constants import (
-    appname, is_macos, is_wayland, supports_primary_selection
+    appname, is_macos, is_wayland, shell_path, supports_primary_selection
 )
+from .options_stub import Options
 from .rgb import Color, to_color
+from .typing import AddressFamily, PopenType, Socket, StartupCtx
 
 BASE = os.path.dirname(os.path.abspath(__file__))
 
 
-def load_shaders(name):
+def expandvars(val: str, env: Dict[str, str] = {}) -> str:
+
+    def sub(m: Match) -> str:
+        key = m.group(1)
+        result = env.get(key)
+        if result is None:
+            result = os.environ.get(key)
+        if result is None:
+            result = m.group()
+        return result
+
+    return re.sub(r'\$\{(\S+?)\}', sub, val)
+
+
+def load_shaders(name: str) -> Tuple[str, str]:
     from .fast_data_types import GLSL_VERSION
     with open(os.path.join(BASE, '{}_vertex.glsl'.format(name))) as f:
         vert = f.read().replace('GLSL_VERSION', str(GLSL_VERSION), 1)
@@ -30,81 +51,94 @@ def load_shaders(name):
     return vert, frag
 
 
-def safe_print(*a, **k):
+def safe_print(*a: Any, **k: Any) -> None:
     with suppress(Exception):
         print(*a, **k)
 
 
-def log_error(*a, **k):
+def log_error(*a: Any, **k: str) -> None:
     from .fast_data_types import log_error_string
     with suppress(Exception):
         msg = k.get('sep', ' ').join(map(str, a)) + k.get('end', '')
         log_error_string(msg.replace('\0', ''))
 
 
-def ceil_int(x):
+def ceil_int(x: float) -> int:
     return int(math.ceil(x))
 
 
-def sanitize_title(x):
+def sanitize_title(x: str) -> str:
     return re.sub(r'\s+', ' ', re.sub(r'[\0-\x19\x80-\x9f]', '', x))
 
 
-def color_as_int(val):
+def color_as_int(val: Tuple[int, int, int]) -> int:
     return val[0] << 16 | val[1] << 8 | val[2]
 
 
-def color_from_int(val):
+def color_from_int(val: int) -> Color:
     return Color((val >> 16) & 0xFF, (val >> 8) & 0xFF, val & 0xFF)
 
 
-def parse_color_set(raw):
+def parse_color_set(raw: str) -> Generator[Tuple[int, Optional[int]], None, None]:
     parts = raw.split(';')
     lp = len(parts)
     if lp % 2 != 0:
         return
-    for c, spec in [parts[i:i + 2] for i in range(0, len(parts), 2)]:
+    for c_, spec in [parts[i:i + 2] for i in range(0, len(parts), 2)]:
         try:
-            c = int(c)
+            c = int(c_)
             if c < 0 or c > 255:
                 continue
             if spec == '?':
                 yield c, None
             else:
-                r, g, b = to_color(spec)
-                yield c, r << 16 | g << 8 | b
+                q = to_color(spec)
+                if q is not None:
+                    r, g, b = q
+                    yield c, r << 16 | g << 8 | b
         except Exception:
             continue
 
 
-def screen_size_function(fd=None):
-    ans = getattr(screen_size_function, 'ans', None)
-    if ans is None:
-        from collections import namedtuple
-        import array
-        import fcntl
-        import termios
-        Size = namedtuple('Size', 'rows cols width height cell_width cell_height')
+class ScreenSize(NamedTuple):
+    rows: int
+    cols: int
+    width: int
+    height: int
+    cell_width: int
+    cell_height: int
+
+
+class ScreenSizeGetter:
+    changed = True
+    Size = ScreenSize
+    ans: Optional[ScreenSize] = None
+
+    def __init__(self, fd: Optional[int]):
         if fd is None:
-            fd = sys.stdout
+            fd = sys.stdout.fileno()
+        self.fd = fd
 
-        def screen_size():
-            if screen_size.changed:
-                buf = array.array('H', [0, 0, 0, 0])
-                fcntl.ioctl(fd, termios.TIOCGWINSZ, buf)
-                rows, cols, width, height = tuple(buf)
-                cell_width, cell_height = width // (cols or 1), height // (rows or 1)
-                screen_size.ans = Size(rows, cols, width, height, cell_width, cell_height)
-                screen_size.changed = False
-            return screen_size.ans
-        screen_size.changed = True
-        screen_size.Size = Size
-        ans = screen_size_function.ans = screen_size
-
-    return ans
+    def __call__(self) -> ScreenSize:
+        if self.changed:
+            import array
+            import fcntl
+            import termios
+            buf = array.array('H', [0, 0, 0, 0])
+            fcntl.ioctl(self.fd, termios.TIOCGWINSZ, cast(bytearray, buf))
+            rows, cols, width, height = tuple(buf)
+            cell_width, cell_height = width // (cols or 1), height // (rows or 1)
+            self.ans = ScreenSize(rows, cols, width, height, cell_width, cell_height)
+            self.changed = False
+        return cast(ScreenSize, self.ans)
 
 
-def fit_image(width, height, pwidth, pheight):
+@lru_cache(maxsize=64)
+def screen_size_function(fd: Optional[int] = None) -> ScreenSizeGetter:
+    return ScreenSizeGetter(fd)
+
+
+def fit_image(width: int, height: int, pwidth: int, pheight: int) -> Tuple[int, int]:
     from math import floor
     if height > pheight:
         corrf = pheight / float(height)
@@ -119,27 +153,25 @@ def fit_image(width, height, pwidth, pheight):
     return int(width), int(height)
 
 
-def set_primary_selection(text):
+def set_primary_selection(text: Union[str, bytes]) -> None:
     if not supports_primary_selection:
         return  # There is no primary selection
-    if isinstance(text, bytes):
-        text = text.decode('utf-8')
-    from kitty.fast_data_types import set_primary_selection
-    set_primary_selection(text)
+    from kitty.fast_data_types import set_primary_selection as s
+    s(text)
 
 
-def get_primary_selection():
+def get_primary_selection() -> str:
     if not supports_primary_selection:
         return ''  # There is no primary selection
-    from kitty.fast_data_types import get_primary_selection
-    return (get_primary_selection() or b'').decode('utf-8', 'replace')
+    from kitty.fast_data_types import get_primary_selection as g
+    return (g() or b'').decode('utf-8', 'replace')
 
 
 def base64_encode(
-    integer,
-    chars=string.ascii_uppercase + string.ascii_lowercase + string.digits +
+    integer: int,
+    chars: str = string.ascii_uppercase + string.ascii_lowercase + string.digits +
     '+/'
-):
+) -> str:
     ans = ''
     while True:
         integer, remainder = divmod(integer, 64)
@@ -149,7 +181,7 @@ def base64_encode(
     return ans
 
 
-def command_for_open(program='default'):
+def command_for_open(program: Union[str, List[str]] = 'default') -> List[str]:
     if isinstance(program, str):
         from .conf.utils import to_cmdline
         program = to_cmdline(program)
@@ -160,22 +192,22 @@ def command_for_open(program='default'):
     return cmd
 
 
-def open_cmd(cmd, arg=None, cwd=None):
+def open_cmd(cmd: Union[Iterable[str], List[str]], arg: Union[None, Iterable[str], str] = None, cwd: Optional[str] = None) -> PopenType:
     import subprocess
     if arg is not None:
         cmd = list(cmd)
-        if isinstance(arg, (list, tuple)):
-            cmd.extend(arg)
-        else:
+        if isinstance(arg, str):
             cmd.append(arg)
-    return subprocess.Popen(cmd, stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, cwd=cwd or None)
+        else:
+            cmd.extend(arg)
+    return subprocess.Popen(tuple(cmd), stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, cwd=cwd or None)
 
 
-def open_url(url, program='default', cwd=None):
+def open_url(url: str, program: Union[str, List[str]] = 'default', cwd: Optional[str] = None) -> PopenType:
     return open_cmd(command_for_open(program), url, cwd=cwd)
 
 
-def detach(fork=True, setsid=True, redirect=True):
+def detach(fork: bool = True, setsid: bool = True, redirect: bool = True) -> None:
     if fork:
         # Detach from the controlling process.
         if os.fork() != 0:
@@ -187,36 +219,36 @@ def detach(fork=True, setsid=True, redirect=True):
         redirect_std_streams(os.devnull)
 
 
-def adjust_line_height(cell_height, val):
+def adjust_line_height(cell_height: int, val: Union[int, float]) -> int:
     if isinstance(val, int):
         return cell_height + val
     return int(cell_height * val)
 
 
-def init_startup_notification_x11(window_handle, startup_id=None):
+def init_startup_notification_x11(window_handle: int, startup_id: Optional[str] = None) -> Optional['StartupCtx']:
     # https://specifications.freedesktop.org/startup-notification-spec/startup-notification-latest.txt
     from kitty.fast_data_types import init_x11_startup_notification
     sid = startup_id or os.environ.pop('DESKTOP_STARTUP_ID', None)  # ensure child processes don't get this env var
     if not sid:
-        return
+        return None
     from .fast_data_types import x11_display
     display = x11_display()
     if not display:
-        return
+        return None
     return init_x11_startup_notification(display, window_handle, sid)
 
 
-def end_startup_notification_x11(ctx):
+def end_startup_notification_x11(ctx: 'StartupCtx') -> None:
     from kitty.fast_data_types import end_x11_startup_notification
     end_x11_startup_notification(ctx)
 
 
-def init_startup_notification(window_handle, startup_id=None):
+def init_startup_notification(window_handle: Optional[int], startup_id: Optional[str] = None) -> Optional['StartupCtx']:
     if is_macos or is_wayland():
-        return
+        return None
     if window_handle is None:
         log_error('Could not perform startup notification as window handle not present')
-        return
+        return None
     try:
         return init_startup_notification_x11(window_handle, startup_id)
     except Exception:
@@ -224,7 +256,7 @@ def init_startup_notification(window_handle, startup_id=None):
         traceback.print_exc()
 
 
-def end_startup_notification(ctx):
+def end_startup_notification(ctx: Optional['StartupCtx']) -> None:
     if not ctx:
         return
     if is_macos or is_wayland():
@@ -238,15 +270,15 @@ def end_startup_notification(ctx):
 
 class startup_notification_handler:
 
-    def __init__(self, do_notify=True, startup_id=None, extra_callback=None):
+    def __init__(self, do_notify: bool = True, startup_id: Optional[str] = None, extra_callback: Optional[Callable] = None):
         self.do_notify = do_notify
         self.startup_id = startup_id
         self.extra_callback = extra_callback
-        self.ctx = None
+        self.ctx: Optional['StartupCtx'] = None
 
-    def __enter__(self):
+    def __enter__(self) -> Callable[[int], None]:
 
-        def pre_show_callback(window_handle):
+        def pre_show_callback(window_handle: int) -> None:
             if self.extra_callback is not None:
                 self.extra_callback(window_handle)
             if self.do_notify:
@@ -254,12 +286,12 @@ class startup_notification_handler:
 
         return pre_show_callback
 
-    def __exit__(self, *a):
+    def __exit__(self, *a: Any) -> None:
         if self.ctx is not None:
             end_startup_notification(self.ctx)
 
 
-def remove_socket_file(s, path=None):
+def remove_socket_file(s: 'Socket', path: Optional[str] = None) -> None:
     with suppress(OSError):
         s.close()
     if path:
@@ -267,7 +299,7 @@ def remove_socket_file(s, path=None):
             os.unlink(path)
 
 
-def unix_socket_paths(name, ext='.lock'):
+def unix_socket_paths(name: str, ext: str = '.lock') -> Generator[str, None, None]:
     import tempfile
     home = os.path.expanduser('~')
     candidates = [tempfile.gettempdir(), home]
@@ -280,7 +312,7 @@ def unix_socket_paths(name, ext='.lock'):
             yield os.path.join(loc, filename)
 
 
-def single_instance_unix(name):
+def single_instance_unix(name: str) -> bool:
     import socket
     for path in unix_socket_paths(name):
         socket_path = path.rpartition('.')[0] + '.sock'
@@ -309,38 +341,47 @@ def single_instance_unix(name):
         s.listen()
         s.set_inheritable(False)
         return True
+    return False
 
 
-def single_instance(group_id=None):
-    import socket
-    name = '{}-ipc-{}'.format(appname, os.geteuid())
-    if group_id:
-        name += '-{}'.format(group_id)
+class SingleInstance:
 
-    s = socket.socket(family=socket.AF_UNIX)
-    # First try with abstract UDS
-    addr = '\0' + name
-    try:
-        s.bind(addr)
-    except OSError as err:
-        if err.errno == errno.ENOENT:
-            return single_instance_unix(name)
-        if err.errno == errno.EADDRINUSE:
-            s.connect(addr)
-            single_instance.socket = s
-            return False
-        raise
-    s.listen()
-    single_instance.socket = s  # prevent garbage collection from closing the socket
-    s.set_inheritable(False)
-    atexit.register(remove_socket_file, s)
-    return True
+    socket: Optional['Socket'] = None
+
+    def __call__(self, group_id: Optional[str] = None) -> bool:
+        import socket
+        name = '{}-ipc-{}'.format(appname, os.geteuid())
+        if group_id:
+            name += '-{}'.format(group_id)
+
+        s = socket.socket(family=socket.AF_UNIX)
+        # First try with abstract UDS
+        addr = '\0' + name
+        try:
+            s.bind(addr)
+        except OSError as err:
+            if err.errno == errno.ENOENT:
+                return single_instance_unix(name)
+            if err.errno == errno.EADDRINUSE:
+                s.connect(addr)
+                self.socket = s
+                return False
+            raise
+        s.listen()
+        self.socket = s  # prevent garbage collection from closing the socket
+        s.set_inheritable(False)
+        atexit.register(remove_socket_file, s)
+        return True
 
 
-def parse_address_spec(spec):
+single_instance = SingleInstance()
+
+
+def parse_address_spec(spec: str) -> Tuple[AddressFamily, Union[Tuple[str, int], str], Optional[str]]:
     import socket
     protocol, rest = spec.split(':', 1)
     socket_path = None
+    address: Union[str, Tuple[str, int]] = ''
     if protocol == 'unix':
         family = socket.AF_UNIX
         address = rest
@@ -357,7 +398,7 @@ def parse_address_spec(spec):
     return family, address, socket_path
 
 
-def write_all(fd, data):
+def write_all(fd: int, data: Union[str, bytes]) -> None:
     if isinstance(data, str):
         data = data.encode('utf-8')
     while data:
@@ -369,23 +410,23 @@ def write_all(fd, data):
 
 class TTYIO:
 
-    def __enter__(self):
+    def __enter__(self) -> 'TTYIO':
         from .fast_data_types import open_tty
         self.tty_fd, self.original_termios = open_tty(True)
         return self
 
-    def __exit__(self, *a):
+    def __exit__(self, *a: Any) -> None:
         from .fast_data_types import close_tty
         close_tty(self.tty_fd, self.original_termios)
 
-    def send(self, data):
+    def send(self, data: Union[str, bytes, Iterable[Union[str, bytes]]]) -> None:
         if isinstance(data, (str, bytes)):
             write_all(self.tty_fd, data)
         else:
             for chunk in data:
                 write_all(self.tty_fd, chunk)
 
-    def recv(self, more_needed, timeout, sz=1):
+    def recv(self, more_needed: Callable[[bytes], bool], timeout: float, sz: int = 1) -> None:
         fd = self.tty_fd
         start_time = monotonic()
         while timeout > monotonic() - start_time:
@@ -396,46 +437,38 @@ class TTYIO:
                 break
 
 
-def natsort_ints(iterable):
+def natsort_ints(iterable: Iterable[str]) -> List[str]:
 
-    def convert(text):
+    def convert(text: str) -> Union[int, str]:
         return int(text) if text.isdigit() else text
 
-    def alphanum_key(key):
+    def alphanum_key(key: str) -> Tuple[Union[int, str], ...]:
         return tuple(map(convert, re.split(r'(\d+)', key)))
 
     return sorted(iterable, key=alphanum_key)
 
 
-def exe_exists(exe):
-    for loc in os.environ.get('PATH', '').split(os.pathsep):
-        if loc and os.access(os.path.join(loc, exe), os.X_OK):
-            return os.path.join(loc, exe)
-    return False
+@lru_cache(maxsize=2)
+def get_editor() -> List[str]:
+    import shlex
+    import shutil
+    for ans in (os.environ.get('VISUAL'), os.environ.get('EDITOR'), 'vim',
+                'nvim', 'vi', 'emacs', 'kak', 'micro', 'nano', 'vis'):
+        if ans and shutil.which(shlex.split(ans)[0]):
+            break
+    else:
+        ans = 'vim'
+    return shlex.split(ans)
 
 
-def get_editor():
-    ans = getattr(get_editor, 'ans', False)
-    if ans is False:
-        import shlex
-        for ans in (os.environ.get('VISUAL'), os.environ.get('EDITOR'), 'vim',
-                    'nvim', 'vi', 'emacs', 'kak', 'micro', 'nano', 'vis'):
-            if ans and exe_exists(shlex.split(ans)[0]):
-                break
-        else:
-            ans = 'vim'
-        ans = shlex.split(ans)
-        get_editor.ans = ans
-    return ans
-
-
-def is_path_in_temp_dir(path):
+def is_path_in_temp_dir(path: str) -> bool:
     if not path:
         return False
 
-    def abspath(x):
+    def abspath(x: Optional[str]) -> str:
         if x:
-            return os.path.abspath(os.path.realpath(x))
+            x = os.path.abspath(os.path.realpath(x))
+        return x or ''
 
     import tempfile
     path = abspath(path)
@@ -446,9 +479,88 @@ def is_path_in_temp_dir(path):
     return False
 
 
-def func_name(f):
+def func_name(f: Any) -> str:
     if hasattr(f, '__name__'):
-        return f.__name__
+        return str(f.__name__)
     if hasattr(f, 'func') and hasattr(f.func, '__name__'):
-        return f.func.__name__
+        return str(f.func.__name__)
     return str(f)
+
+
+def resolved_shell(opts: Optional[Options] = None) -> List[str]:
+    q: str = getattr(opts, 'shell', '.')
+    if q == '.':
+        ans = [shell_path]
+    else:
+        import shlex
+        ans = shlex.split(q)
+    return ans
+
+
+def read_shell_environment(opts: Optional[Options] = None) -> Dict[str, str]:
+    ans: Optional[Dict[str, str]] = getattr(read_shell_environment, 'ans', None)
+    if ans is None:
+        from .child import openpty, remove_blocking
+        ans = {}
+        setattr(read_shell_environment, 'ans', ans)
+        import subprocess
+        shell = resolved_shell(opts)
+        master, slave = openpty()
+        remove_blocking(master)
+        try:
+            p = subprocess.Popen(shell + ['-l', '-c', 'env'], stdout=slave, stdin=slave, stderr=slave, start_new_session=True, close_fds=True)
+        except FileNotFoundError:
+            log_error('Could not find shell to read environment')
+            return ans
+        with os.fdopen(master, 'rb') as stdout, os.fdopen(slave, 'wb'):
+            raw = b''
+            from subprocess import TimeoutExpired
+            from time import monotonic
+            start_time = monotonic()
+            while monotonic() - start_time < 1.5:
+                try:
+                    ret: Optional[int] = p.wait(0.01)
+                except TimeoutExpired:
+                    ret = None
+                with suppress(Exception):
+                    raw += stdout.read()
+                if ret is not None:
+                    break
+            if cast(Optional[int], p.returncode) is None:
+                log_error('Timed out waiting for shell to quit while reading shell environment')
+                p.kill()
+            elif p.returncode == 0:
+                while True:
+                    try:
+                        x = stdout.read()
+                    except Exception:
+                        break
+                    if not x:
+                        break
+                    raw += x
+                draw = raw.decode('utf-8', 'replace')
+                for line in draw.splitlines():
+                    k, v = line.partition('=')[::2]
+                    if k and v:
+                        ans[k] = v
+            else:
+                log_error('Failed to run shell to read its environment')
+    return ans
+
+
+def parse_uri_list(text: str) -> Generator[str, None, None]:
+    ' Get paths from file:// URLs '
+    from urllib.parse import urlparse, unquote
+    for line in text.splitlines():
+        if not line or line.startswith('#'):
+            continue
+        if not line.startswith('file://'):
+            yield line
+            continue
+        try:
+            purl = urlparse(line, allow_fragments=False)
+        except Exception:
+            yield line
+            continue
+        if purl.path:
+            yield unquote(purl.path)
